#pragma once
#include <ROOT/RDataFrame.hxx>
#include <fstream>
#include <TCanvas.h>
#include <TLorentzVector.h>
#include <Math/LorentzVector.h>
#include <ROOT/RDF/RSampleInfo.hxx>
#include <Math/GenVector/VectorUtil.h>

ROOT::VecOps::RVec<ROOT::Math::PxPyPzEVector> p4ToGenVecP4(const ROOT::VecOps::RVec<TLorentzVector>& vec){
    ROOT::VecOps::RVec<ROOT::Math::PxPyPzEVector> ret;
    ret.reserve(vec.size());
    for (auto entry : vec){
        ret.emplace_back(entry);
    }
    return ret;
}

ROOT::Math::PxPyPzEVector p4ToGenVecP4Single(const TLorentzVector& p4){
    return ROOT::Math::PxPyPzEVector(p4);
}

double getMetaReweight(const uint slot, const ROOT::RDF::RSampleInfo& sampleinfo){
    if (!sampleinfo.GetI("isMC")) return 1.0;
    double k_fac        = sampleinfo.GetD("kFactor");
    double gen_filt_eff = sampleinfo.GetD("genFiltEff");
    double xs_in_fb     = sampleinfo.GetD("crossSection") * 1e3; //pb to fb
    double sum_w_mc     = sampleinfo.GetD("sum_MC_weighted_events");
    return k_fac * gen_filt_eff * xs_in_fb / sum_w_mc;
}

int getMetaDSID(const uint slot, const ROOT::RDF::RSampleInfo& sampleinfo){
    int dsid = sampleinfo.GetI("dataset_number");
    assert (dsid > 600000);
    return dsid;
}

template<typename T>
std::vector<T> RVecToVec(const ROOT::VecOps::RVec<T>& rvec){
    std::vector<T> ret(rvec.begin(), rvec.end());
    return ret;
}

double getSignedDPhi(double v1phi,  double v2phi)
{
    double ret;
    if(std::abs(v1phi - v2phi) > M_PI)
        ret = ((double)(v1phi > 0) -  (double)(v1phi <= 0)) * (2 * M_PI - std::abs(v1phi - v2phi));
    else
        ret = v2phi - v1phi;
    return ret;
}

double getOmega(ROOT::Math::PxPyPzEVector tau_p4, ROOT::Math::PxPyPzEVector mu_p4, ROOT::Math::PxPyPzEVector met_p4){
    auto tau_phi = tau_p4.phi(); auto mu_phi = mu_p4.phi(); auto met_phi = met_p4.phi();
    auto dphi_taumu  = getSignedDPhi(tau_phi,  mu_phi);
    auto dphi_mumet  = getSignedDPhi(mu_phi,   met_phi);
    auto dphi_taumet = getSignedDPhi(tau_phi,  met_phi);
    bool met_inside_taumu = (dphi_taumu * dphi_taumet > 0) && (std::abs(dphi_taumu) > std::abs(dphi_taumet));
    bool met_close_to_tau = std::abs(dphi_taumet) < M_PI_2 && std::abs(dphi_taumet) < std::abs(dphi_mumet);
    bool met_close_to_mu  = std::abs(dphi_mumet)  < M_PI_2 && std::abs(dphi_mumet) < std::abs(dphi_taumet);
    bool met_pos_pass = met_inside_taumu || met_close_to_tau || met_close_to_mu;
    auto omega = -100.0;
    if(!met_inside_taumu && met_close_to_mu && dphi_taumet * dphi_taumu < 0) {
        if(dphi_taumu < 0) {
            omega = (dphi_taumet - 2.0 * M_PI) / dphi_taumu;
        }
        else if (dphi_taumet < 0) {
            omega = (dphi_taumet + M_PI * 2.0) / dphi_taumu;
        }
    }
    else {
        omega = dphi_taumet / dphi_taumu;
    }
    return omega;
}

double getOmegaNoScale(ROOT::Math::PxPyPzEVector tau_p4, ROOT::Math::PxPyPzEVector mu_p4, ROOT::Math::PxPyPzEVector met_p4){
    auto tau_phi = tau_p4.phi(); auto mu_phi = mu_p4.phi(); auto met_phi = met_p4.phi();
    auto dphi_taumu  = getSignedDPhi(tau_phi,  mu_phi);
    auto dphi_mumet  = getSignedDPhi(mu_phi,   met_phi);
    auto dphi_taumet = getSignedDPhi(tau_phi,  met_phi);
    bool met_inside_taumu = (dphi_taumu * dphi_taumet > 0) && (std::abs(dphi_taumu) > std::abs(dphi_taumet));
    bool met_close_to_tau = std::abs(dphi_taumet) < M_PI_2 && std::abs(dphi_taumet) < std::abs(dphi_mumet);
    bool met_close_to_mu  = std::abs(dphi_mumet)  < M_PI_2 && std::abs(dphi_mumet) < std::abs(dphi_taumet);
    bool met_pos_pass = met_inside_taumu || met_close_to_tau || met_close_to_mu;
    auto omega = -100.0;
    if(!met_inside_taumu && met_close_to_mu && dphi_taumet * dphi_taumu < 0) {
        if(dphi_taumu < 0) {
            omega = (dphi_taumet - 2.0 * M_PI) / dphi_taumu;
        }
        else if (dphi_taumet < 0) {
            omega = (dphi_taumet + M_PI * 2.0) / dphi_taumu;
        }
    }
    else {
        omega = dphi_taumet / dphi_taumu;
    }
    return omega * std::abs(dphi_taumu);
}

ROOT::Math::PxPyPzEVector colinP4(ROOT::Math::PxPyPzEVector tau_p4, ROOT::Math::PxPyPzEVector mu_p4, ROOT::Math::PxPyPzEVector met_p4){
    auto dphi_taumu  = getSignedDPhi(tau_p4.phi(),  mu_p4.phi());
    auto dphi_mumet  = getSignedDPhi(mu_p4.phi(),   met_p4.phi());
    auto dphi_taumet = getSignedDPhi(tau_p4.phi(),  met_p4.phi());
    bool met_inside_taumu = (dphi_taumu * dphi_taumet > 0) && (std::abs(dphi_taumu) > std::abs(dphi_taumet));
    bool met_close_to_tau = std::abs(dphi_taumet) < M_PI_2 && std::abs(dphi_taumet) < std::abs(dphi_mumet);
    bool met_close_to_mu  = std::abs(dphi_mumet)  < M_PI_2 && std::abs(dphi_mumet) < std::abs(dphi_taumet);
    bool met_pos_pass = met_inside_taumu || met_close_to_tau || met_close_to_mu;
    
    ROOT::Math::PtEtaPhiMVector vh_p4(0,0,0,0), vl_p4(0,0,0,0);
    if(!met_pos_pass) 
        return ROOT::Math::PxPyPzEVector(0,0,0,0);
    if (!met_inside_taumu && met_close_to_tau) 
    {
        vh_p4.SetPt(met_p4.pt() * std::cos(std::abs(dphi_taumet)));
        vh_p4.SetEta(tau_p4.eta());
        vh_p4.SetPhi(tau_p4.phi());
        vh_p4.SetM(0);
    }
    else if (!met_inside_taumu && met_close_to_mu)
    {
        vl_p4.SetPt(met_p4.pt() * std::cos(std::abs(dphi_mumet)));
        vl_p4.SetEta(mu_p4.eta());
        vl_p4.SetPhi(mu_p4.phi());
        vl_p4.SetM(0);
    }
    else if (met_inside_taumu)
    {
        vh_p4.SetPt(met_p4.pt() * std::cos(std::abs(dphi_taumet)) - met_p4.pt() * std::sin(std::abs(dphi_taumet)) * (1/std::tan(std::abs(dphi_taumu))));
        vh_p4.SetEta(tau_p4.eta());
        vh_p4.SetPhi(tau_p4.phi());
        vh_p4.SetM(0);

        vl_p4.SetPt(met_p4.pt() * std::sin(std::abs(dphi_taumet)) / std::sin(std::abs(dphi_taumu)));
        vl_p4.SetEta(mu_p4.eta());
        vl_p4.SetPhi(mu_p4.phi());
        vl_p4.SetM(0);
    }
    auto p4_col_alt = tau_p4 + vh_p4 + vl_p4 + mu_p4;
    auto p4_ret = ROOT::Math::PxPyPzEVector(p4_col_alt.Px(), p4_col_alt.Py(), p4_col_alt.Pz(), p4_col_alt.E());
    return p4_ret;
}

double getMinDR(const ROOT::Math::PxPyPzEVector& p0, const ROOT::Math::PxPyPzEVector& p1, const ROOT::Math::PxPyPzEVector& p2){
    double dR01 = ROOT::Math::VectorUtil::DeltaR(p0, p1);
    double dR02 = ROOT::Math::VectorUtil::DeltaR(p0, p2);
    return dR01 > dR02 ? dR02 : dR01;
}

uint getMinDRPaticle(const ROOT::Math::PxPyPzEVector& p0, const ROOT::VecOps::RVec<ROOT::Math::PxPyPzEVector>& p_targets){
    uint ret;
    double min_dR = 1000;
    for (uint i = 0; i < p_targets.size(); i++){
        auto dr_now = ROOT::Math::VectorUtil::DeltaR(p0, p_targets[i]);
        if (dr_now < min_dR){
            min_dR = dr_now;
            ret = i;
        }
    }
    return ret;
}

template<typename T1, typename T2>
ROOT::Math::PxPyPzEVector boostVec(ROOT::Math::LorentzVector<T1> vec, ROOT::Math::LorentzVector<T2> ref){
    auto ref_boost_vec = ref.BoostToCM();
    auto boost = ROOT::Math::Boost(ref_boost_vec);
    return boost(vec);
}

ROOT::RVec<bool> highestScoreIdx(const ROOT::RVec<float>& scores){
    uint max_idx = 0;
    for (uint idx = 0; idx < scores.size(); idx++){
        if (scores[idx] > scores[max_idx]){
            max_idx = idx;
        }
    }
    auto ret = ROOT::RVec<bool>(scores.size());
    ret[max_idx] = true;
    return ret;
}

template <typename T_df>
ROOT::RDF::RNode getFinalWeight(T_df df, double luminosity){
    return df
        .DefinePerSample("tmp_weight_sf", getMetaReweight)
        .Define("final_weight", [=](double tmp_weight, double weight_mc) { return tmp_weight * weight_mc * luminosity; }, {"tmp_weight_sf", "weight_mc"})
        .DefinePerSample("sample_dsid", getMetaDSID)
    ;
}

template <typename T_df>
ROOT::RDF::RNode analysisMain(T_df df){
    using namespace ROOT::VecOps; 
    using namespace ROOT::Math; 

    auto ret = df
        .Define("best_taus_muonrm",  highestScoreIdx, {"taus_muonrm_jet_deepset_score_trans"})
        .Alias ("good_taus_muonrm", "best_taus_muonrm")
        .Define("leading_tau_rnn",  [](const RVec<float>& scores,           const RVec<bool>& id_flags){return scores[id_flags].at(0);}, {"taus_muonrm_jet_rnn_score_trans", "good_taus_muonrm"})
        .Define("leading_tau_dpst", [](const RVec<float>& scores,           const RVec<bool>& id_flags){return scores[id_flags].at(0);}, {"taus_muonrm_jet_deepset_score_trans", "good_taus_muonrm"})
        .Define("leading_tau_p4",   [](const RVec<TLorentzVector>& p4s_leg, const RVec<bool>& id_flags){return p4ToGenVecP4(p4s_leg)[id_flags].at(0);},   {"taus_muonrm_p4", "good_taus_muonrm"})
        .Define("leading_tau_q",    [](const RVec<float>& qs,               const RVec<bool>& id_flags){return qs[id_flags].at(0);},                                    {"taus_muonrm_q", "good_taus_muonrm"})
        .Define("corr_mu_p4",       [](const RVec<TLorentzVector>& p4s_leg, const RVec<bool>&  id_flags){return p4ToGenVecP4(p4s_leg)[id_flags].at(0);},    {"muons_rm_p4", "good_taus_muonrm"})
        .Define("corr_mu_q",        [](const RVec<float>& qs,               const RVec<bool>& id_flags){return qs[id_flags].at(0);},                                     {"muons_rm_q", "good_taus_muonrm"})
        .Define("mutau_vis_p4",     [](const PxPyPzEVector& p4_1,           const PxPyPzEVector& p4_2){ return p4_1 + p4_2; },                                      {"leading_tau_p4", "corr_mu_p4"})
        .Define("event_contain_b",  [](const RVec<float>& b_veto_flag){ return Sum(b_veto_flag) > 0; },                                            {"jets_b_tagged_DL1dv01_FixedCutBEff_85"})
        .Define("met_genvec_p4",    p4ToGenVecP4Single,                                                                                     {"met_p4"})
        .Define("mutau_colin_p4",   colinP4,                                                                                                {"leading_tau_p4", "corr_mu_p4", "met_genvec_p4"})
        .Filter(                    [](const ROOT::Math::PxPyPzEVector& col_p4){return col_p4.pt() > 0;},                                        {"mutau_colin_p4"}, "colin_pt_cut")
        .Define("omega",            getOmega,                                                                                               {"leading_tau_p4", "corr_mu_p4", "met_genvec_p4"})
        .Define("dR_mutau",         [](const PxPyPzEVector& p4_1, const PxPyPzEVector& p4_2){return ROOT::Math::VectorUtil::DeltaR(p4_1, p4_2);},         {"corr_mu_p4", "leading_tau_p4"})
        .Define("omega_noscale",    getOmegaNoScale,                                                                                                    {"leading_tau_p4", "corr_mu_p4", "met_genvec_p4"})
    ;
    return ret;
}

std::vector<double> sumRVecD(ROOT::RDF::RNode df, std::string colname){
    auto size = df.Range(0, 1).Take<ROOT::RVecD>(colname).GetValue()[0].size();
    auto ret = df.Reduce([](ROOT::RVecD v1, ROOT::RVecD v2){return v1 + v2;}, colname, ROOT::RVecD(size)).GetValue();
    return RVecToVec(ret);
}

template <typename T_df>
ROOT::RDF::RNode filterTriggers(T_df df, std::vector<std::string> triggers){
    auto col_names = df.GetColumnNames();
    std::string tgr_ftr_string = "return ";
    for (auto tgr_name : triggers){
        tgr_ftr_string += tgr_name + " || ";
    }
    tgr_ftr_string.erase(tgr_ftr_string.size() - 4, 4);
    return df.Filter(tgr_ftr_string, "filter_tgr_decision");;
}