import ROOT
import json
import atlasplots as aplt
from const import *


def fillHistos(dfs_dict, hist_expr, weight, hist_model, yoi, pre_filter_expr = 'true', post_filter_expr = 'true'):
    ret = {}
    for years in yoi: 
        ret[years] = {}
        for key, val in dfs_dict[years].items():
            ret[years][key] = val\
                .Filter(pre_filter_expr)\
                .Define("var_to_be_fill", hist_expr)\
                .Filter(post_filter_expr)\
                .Histo1D(hist_model, "var_to_be_fill", weight)
    return ret

def fillHistos2D(dfs_dict, hist_expr1, hist_expr2, weight, hist_model, yoi, pre_filter_expr = 'true', post_filter_expr = 'true'):
    ret = {}
    for years in yoi: 
        ret[years] = {}
        for key, val in dfs_dict[years].items():
            ret[years][key] = val\
                .Filter(pre_filter_expr)\
                .Define("var1_to_be_fill", hist_expr1)\
                .Define("var2_to_be_fill", hist_expr2)\
                .Filter(post_filter_expr)\
                .Histo2D(hist_model, "var1_to_be_fill", "var2_to_be_fill", weight)
    return ret

def FromSpecSingle(json_file_path):
    ret = {}
    for sample_name, vals in json.load(open(json_file_path))['samples'].items():
        rmd = ROOT.RDF.Experimental.RMetaData()
        dsid = vals['metadata']['dataset_number']
        for key, val in vals['metadata'].items():
            rmd.Add(key, val)
        spec  = ROOT.RDF.Experimental.RDatasetSpec()
        rdg = ROOT.RDF.Experimental.RSample(sample_name, vals['trees'], vals['files'], rmd)
        spec.AddSample(rdg)
        ret[str(dsid)] = ROOT.RDataFrame(spec)
    return ret

def combineHistoByChannel(hs_dict_by_id, channel, color, yoi):
    to_clone = next(iter((next(iter(hs_dict_by_id.values()))).values()))
    ret_h = to_clone.Clone(channel)
    ret_h.Reset("M")
    for years in yoi :
        dsids = sample_dicts[years][channel]
        for dsid in dsids:
            if hs_dict_by_id[years].get(dsid) is not None:
                ret_h.Add(hs_dict_by_id[years][dsid].GetValue())
    ret_h.SetLineColor(color)
    ret_h.SetMarkerColor(color)
    ret_h.SetFillColorAlpha(color, 1.0)
    ret_h.SetFillStyle(1001)
    return ret_h

def makeStack(hs_dict_by_id, yoi, coi):
    all_tasks = [hs_dict_by_id[ayear].get(adsid) for achannel in coi for ayear in yoi for adsid in sample_dicts[ayear][achannel]]
    all_tasks = [task for task in all_tasks if task is not None]
    #parallel
    ROOT.RDF.RunGraphs(all_tasks)
    hs_dict_stack = {}
    ret_stack = ROOT.THStack()
    ret_data_h = None
    for channel in coi:
        hs_dict_stack[channel] = combineHistoByChannel(hs_dict_by_id, channel, color_dict[channel], yoi)
    for channel, h in hs_dict_stack.items():
        if channel == 'data':
            ret_data_h = h
            ret_data_h.SetLineColor(color_dict[channel])
            ret_data_h.SetMarkerColor(color_dict[channel])
            ret_data_h.SetMarkerStyle(8)
            ret_data_h.SetMarkerSize(1)
            ret_data_h.SetLineColor(color_dict[channel])
        else:
            ret_stack.Add(h)
    return ret_stack, ret_data_h

def makeRatioPlt(hs_dict_by_id, x_label, yoi = years_of_interest, coi = physics_sample_list, text = "#sqrt{s} = 13 TeV, 139 fb^{-1}"):
    hs_mc, h_data = makeStack(hs_dict_by_id, yoi=yoi, coi = coi)
    if h_data is not None:
        fig, (ax1, ax2) = aplt.ratio_plot(name=x_label, figsize=(800, 800), hspace=0.05)
        ax1.cd()
        # stack ploting
        ax1.plot(hs_mc)
        mc_err_band = aplt.root_helpers.hist_to_graph(
            hs_mc.GetStack().Last(),
            show_bin_width=True
        )
        ax1.plot(mc_err_band, "2", fillcolor=1, fillstyle=3254, linewidth=0)

        # data graph
        data_graph = aplt.root_helpers.hist_to_graph(h_data)
        ax1.plot(data_graph, "P")

        #legend
        legend = ax1.legend(
            loc      = (0.68, 0.3, 1 - ROOT.gPad.GetRightMargin() - 0.03, 1 - ROOT.gPad.GetTopMargin() - 0.03), 
            textsize = 22,
        )
        for mchist in hs_mc.GetHists():
            legend.AddEntry(mchist, mchist.GetName(), 'F')
        legend.AddEntry(data_graph, 'Data', 'EP')
        legend.AddEntry(mc_err_band, 'MC Stat. Unc.', 'F')
        #misc
        ax1.add_margins(top=0.16)
        ax1.set_ylabel("Event yield")
        aplt.atlas_label(text="Internal", loc="upper left")
        ax1.text(0.2, 0.84, text, size=22, align=13)


        #ratio part
        ax2.cd()
        # Draw line at y=1 in ratio panel
        line = ROOT.TLine(ax1.get_xlim()[0], 1, ax1.get_xlim()[1], 1)
        ax2.plot(line)
        # Plot the relative error on the ratio axes
        err_band_ratio = aplt.root_helpers.hist_to_graph(
            hs_mc.GetStack().Last(),
            show_bin_width=True,
            norm=True
        )
        ax2.plot(err_band_ratio, "2", fillcolor=1, fillstyle=3254)
        # Calculate and draw the ratio
        ratio_hist = h_data.Clone("ratio_hist")
        ratio_hist.Divide(hs_mc.GetStack().Last())
        ratio_graph = aplt.root_helpers.hist_to_graph(ratio_hist)
        ax2.plot(ratio_graph, "P0")
        ax2.set_xlabel(x_label)
        ax2.frame.GetXaxis().SetTitleOffset(1.4)
        ax2.set_ylabel("Data / MC", loc="centre")
        ax2.set_ylim(0, 3)
        # ax2.set_yscale("log")
        ax2.draw_arrows_outside_range(ratio_graph)
        return fig, (ax1, ax2), (hs_mc, mc_err_band, data_graph, line, err_band_ratio, ratio_graph)
    else:
        fig, ax1 = aplt.subplots(1,1, name=x_label, figsize=(800, 800))
        ax1.cd()
        # stack ploting
        ax1.plot(hs_mc)
        mc_err_band = aplt.root_helpers.hist_to_graph(
            hs_mc.GetStack().Last(),
            show_bin_width=True
        )
        ax1.plot(mc_err_band, "2", fillcolor=1, fillstyle=3254, linewidth=0)
        legend = ax1.legend(
            loc      = (0.68, 0.3, 1 - ROOT.gPad.GetRightMargin() - 0.03, 1 - ROOT.gPad.GetTopMargin() - 0.03), 
            textsize = 22,
        )
        for mchist in hs_mc.GetHists():
            legend.AddEntry(mchist, mchist.GetName(), 'F')
        legend.AddEntry(mc_err_band, 'MC Stat. Unc.', 'F')
        #misc
        ax1.add_margins(top=0.16)
        ax1.set_ylabel("Event yield")
        aplt.atlas_label(text="Internal", loc="upper left")
        ax1.text(0.2, 0.84, text, size=22, align=13)
        ax1.set_xlabel(x_label)
        return fig, (ax1, None), (hs_mc, mc_err_band)